#include "Actors/MujocoVolumeActor.h"
#include "Misc/MujocoXmlGenerator.h"
#include "Components/MujocoBodyComponent.h"
#include "Components/MujocoGeomComponent.h"
#include "Components/MujocoSiteComponent.h"
#include "Components/MujocoJointComponent.h"
#include "Components/MujocoActuatorComponent.h"
#include "Components/MujocoEqualityComponent.h"
#include "Components/MujocoTendonComponent.h"
#include "array"
#include "LuckyMujoco.h"
#include "fstream"
#include "filesystem"

AMujocoVolumeActor::AMujocoVolumeActor()
{
	PrimaryActorTick.bCanEverTick = true;
	PrimaryActorTick.bStartWithTickEnabled = true;
	
	Options = CreateDefaultSubobject<UMujocoExportOptions>(TEXT("Options"));

	SpriteComponent = CreateDefaultSubobject<UBillboardComponent>(TEXT("Sprite"));
	RootComponent = SpriteComponent;

	if (SpriteComponent)
	{
		SpriteComponent->bHiddenInGame = true;
		SpriteComponent->bIsScreenSizeScaled = true;
#if WITH_EDITORONLY_DATA
		SpriteComponent->SpriteInfo.Category = TEXT("Mujoco");
		SpriteComponent->SpriteInfo.DisplayName = FText::FromString(TEXT("Mujoco Volume Actor"));
#endif
	}
}

template <typename ComponentType> int32 GetMujocoObjectType();

template <> int32 GetMujocoObjectType<UMujocoBodyComponent>()
{
	return mjOBJ_BODY;
}

template <> int32 GetMujocoObjectType<UMujocoGeomComponent>()
{
	return mjOBJ_GEOM;
}

template <> int32 GetMujocoObjectType<UMujocoSiteComponent>()
{
	return mjOBJ_SITE;
}

template <> int32 GetMujocoObjectType<UMujocoJointComponent>()
{
	return mjOBJ_JOINT;
}

template <> int32 GetMujocoObjectType<UMujocoActuatorComponent>()
{
	return mjOBJ_ACTUATOR;
}

template <> int32 GetMujocoObjectType<UMujocoEqualityComponent>()
{
	return mjOBJ_EQUALITY;
}

template <> int32 GetMujocoObjectType<UMujocoTendonComponent>()
{
	return mjOBJ_TENDON;
}

template <typename ComponentType> void AMujocoVolumeActor::AssignComponentsToArray(UWorld* World, TArray<TSoftObjectPtr<ComponentType>>& ComponentArray)
{
	TMap<int32, TSoftObjectPtr<ComponentType>> ComponentMap;

	int32 MaxIndex = 0;
	for (AActor* Actor : World->GetCurrentLevel()->Actors)
	{
		if (!Actor || Actor->IsPendingKillPending())
		{
			continue;
		}

		TInlineComponentArray<ComponentType*> Components;
		Actor->GetComponents<ComponentType>(Components);
		for (ComponentType* Component : Components)
		{
			std::string ComponentName = TCHAR_TO_ANSI(*Component->GetName());
			int32 MujocoObjectType = GetMujocoObjectType<ComponentType>();
			Component->MujocoID = mj_name2id(MujocoModel.Get(), MujocoObjectType, ComponentName.c_str());
			if (Component->MujocoID >= 0)
			{
				ComponentMap.Add(Component->MujocoID, Component);
				MaxIndex = FMath::Max(MaxIndex, Component->MujocoID);
			}
			else
			{
				UE_LOG(LogMujoco, Warning, TEXT("Failed to assign MujocoID for component: %s"), *Component->GetName());
			}
		}
	}

	ComponentArray.SetNum(MaxIndex + 1);

	for (auto& Pair : ComponentMap)
	{
		ComponentArray[Pair.Key] = Pair.Value;
	}
}

void AMujocoVolumeActor::InitializeMujoco()
{
	if (!Options)
	{
		return;
	}

	if (auto* World = GetWorld())
	{
		//GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Green, TEXT("Generating Mujoco model..."));
		OnMujocoCompileBegin.Broadcast();

		ULevel* Level = World->GetCurrentLevel();
		FString WorldUniqueId = Level->GetOutermost()->GetName();
		WorldUniqueId = FPaths::GetBaseFilename(WorldUniqueId);
		FString SavePath = FPaths::ProjectSavedDir() / TEXT("Mujoco") / WorldUniqueId;						
		SavePath = FPaths::ConvertRelativePathToFull(SavePath);

		IFileManager::Get().MakeDirectory(*SavePath, true);
		FString ExportFilename = SavePath / (Options->ModelName + TEXT(".xml"));

		TArray<UObject*> Objects;

		for (AActor* Actor : World->GetCurrentLevel()->Actors)
		{
			if (!Actor || Actor->IsPendingKillPending())
			{
				continue;
			}

			TInlineComponentArray<UMujocoBodyComponent*> MujocoBodyComponents;
			Actor->GetComponents<UMujocoBodyComponent>(MujocoBodyComponents);
			if (MujocoBodyComponents.Num() == 0)
			{
				continue;
			}
			Objects.Add(Actor);
		}

		TUniquePtr<FMujocoXmlGenerator> Generator = MakeUnique<FMujocoXmlGenerator>();
		TUniquePtr<tinyxml2::XMLDocument> Doc = Generator->GenerateMujocoXml(Options, Objects, ExportFilename);

		FString XmlString;
		tinyxml2::XMLPrinter Printer;
		Doc->Print(&Printer);

		FString PrinterString = UTF8_TO_TCHAR(Printer.CStr());		

		if (!FFileHelper::SaveStringToFile(PrinterString, *ExportFilename))
		{
			//GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, TEXT("Failed to save Mujoco model!"));
			OnMujocoCompileError.Broadcast(FString::Printf(TEXT("Failed to save Mujoco model: %s"), *ExportFilename));
			return;
		}

		std::array<char, 1024> ErrMsg{};
		MujocoModel = MakeMujocoModelPtr(mj_loadXML(TCHAR_TO_ANSI(*ExportFilename), nullptr, ErrMsg.data(), ErrMsg.size()));
		if (!MujocoModel)
		{
			//GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, FString::Printf(TEXT("Failed to load Mujoco model: %s"), UTF8_TO_TCHAR(ErrMsg.data())));
			OnMujocoCompileError.Broadcast(FString::Printf(TEXT("Failed to load Mujoco model: %s"), UTF8_TO_TCHAR(ErrMsg.data())));
			return;
		}

		MujocoData = MakeMujocoDataPtr(MujocoModel.Get());
		if (!MujocoData)
		{
			//GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Red, TEXT("Failed to create Mujoco data!"));
			OnMujocoCompileError.Broadcast(TEXT("Failed to create Mujoco data!"));
			return;
		}

		AssignComponentsToArray(World, BodyComponents);
		AssignComponentsToArray(World, GeomComponents);
		AssignComponentsToArray(World, SiteComponents);
		AssignComponentsToArray(World, JointComponents);
		AssignComponentsToArray(World, ActuatorComponents);
		AssignComponentsToArray(World, EqualityComponents);
		AssignComponentsToArray(World, TendonComponents);

		mj_resetData(MujocoModel.Get(), MujocoData.Get());

		//GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Green, TEXT("Mujoco model generated!"));
		OnMujocoCompileSuccess.Broadcast();
	}
}

void AMujocoVolumeActor::SetActuatorValue(const FString& ActuatorName, double Value)
{
	if (MujocoModel)
	{
		int32 ActuatorID = mj_name2id(MujocoModel.Get(), mjOBJ_ACTUATOR, TCHAR_TO_ANSI(*ActuatorName));
		if (ActuatorID >= 0)
		{
			double MinRange = MujocoModel->actuator_ctrlrange[2 * ActuatorID];
			double MaxRange = MujocoModel->actuator_ctrlrange[2 * ActuatorID + 1];

			Value = FMath::Clamp(Value, MinRange, MaxRange);

			MujocoData->ctrl[ActuatorID] = Value;
		}
		else
		{
			UE_LOG(LogMujoco, Warning, TEXT("Failed to set actuator value for: %s"), *ActuatorName);
		}
	}
}

void AMujocoVolumeActor::SetActuatorValueByIndex(int32 ActuatorIndex, double Value)
{
	if (MujocoModel)
	{		
		if (ActuatorIndex >= 0 && ActuatorIndex < MujocoModel->nu)
		{
			double MinRange = MujocoModel->actuator_ctrlrange[2 * ActuatorIndex];
			double MaxRange = MujocoModel->actuator_ctrlrange[2 * ActuatorIndex + 1];

			Value = FMath::Clamp(Value, MinRange, MaxRange);

			MujocoData->ctrl[ActuatorIndex] = Value;
		}
		else
		{
			UE_LOG(LogMujoco, Warning, TEXT("Failed to set actuator value for index: %d"), ActuatorIndex);
		}
	}
}

double AMujocoVolumeActor::GetActuatorValue(const FString& ActuatorName) const 
{
	if (MujocoModel)
	{
		int32 ActuatorID = mj_name2id(MujocoModel.Get(), mjOBJ_ACTUATOR, TCHAR_TO_ANSI(*ActuatorName));
		if (ActuatorID >= 0)
		{
			return MujocoData->ctrl[ActuatorID];
		}
		else
		{
			UE_LOG(LogMujoco, Warning, TEXT("Failed to get actuator value for: %s"), *ActuatorName);
		}
	}
	return 0.0;
}

double AMujocoVolumeActor::GetActuatorValueByIndex(int32 ActuatorIndex) const
{
	if (MujocoModel)
	{
		if (ActuatorIndex >= 0 && ActuatorIndex < MujocoModel->nu)
		{
			return MujocoData->ctrl[ActuatorIndex];
		}
		else
		{
			UE_LOG(LogMujoco, Warning, TEXT("Failed to get actuator value for index: %d"), ActuatorIndex);
		}
	}
	return 0.0;
}

FVector2D AMujocoVolumeActor::GetActuatorRange(const FString& ActuatorName) const 
{
	if (MujocoModel)
	{
		int32 ActuatorID = mj_name2id(MujocoModel.Get(), mjOBJ_ACTUATOR, TCHAR_TO_ANSI(*ActuatorName));
		if (ActuatorID >= 0)
		{
			return FVector2D(MujocoModel->actuator_ctrlrange[2 * ActuatorID], MujocoModel->actuator_ctrlrange[2 * ActuatorID + 1]);
		}
		else
		{
			UE_LOG(LogMujoco, Warning, TEXT("Failed to get actuator range for: %s"), *ActuatorName);
		}
	}
	return FVector2D(0.0, 0.0);
}

FVector2D AMujocoVolumeActor::GetActuatorRangeByIndex(int32 ActuatorIndex) const 
{
	if (MujocoModel)
	{
		if (ActuatorIndex >= 0 && ActuatorIndex < MujocoModel->nu)
		{
			return FVector2D(MujocoModel->actuator_ctrlrange[2 * ActuatorIndex], MujocoModel->actuator_ctrlrange[2 * ActuatorIndex + 1]);
		}
		else
		{
			UE_LOG(LogMujoco, Warning, TEXT("Failed to get actuator range for index: %d"), ActuatorIndex);
		}
	}
	return FVector2D(0.0, 0.0);
}

void AMujocoVolumeActor::SetJointValue(const FString& JointName, double Value) 
{
	if (MujocoModel)
	{
		int32 JointID = mj_name2id(MujocoModel.Get(), mjOBJ_JOINT, TCHAR_TO_ANSI(*JointName));
		if (JointID >= 0)
		{
			double MinRange = MujocoModel->jnt_range[2 * JointID];
			double MaxRange = MujocoModel->jnt_range[2 * JointID + 1];

			Value = FMath::Clamp(Value, MinRange, MaxRange);

			MujocoData->qpos[MujocoModel->jnt_qposadr[JointID]] = Value;
		}
		else
		{
			UE_LOG(LogMujoco, Warning, TEXT("Failed to set joint value for: %s"), *JointName);
		}
	}
}

void AMujocoVolumeActor::SetJointValueByIndex(int32 JointIndex, double Value) 
{
	if (MujocoModel)
	{
		if (JointIndex >= 0 && JointIndex < MujocoModel->njnt)
		{
			double MinRange = MujocoModel->jnt_range[2 * JointIndex];
			double MaxRange = MujocoModel->jnt_range[2 * JointIndex + 1];

			Value = FMath::Clamp(Value, MinRange, MaxRange);

			MujocoData->qpos[MujocoModel->jnt_qposadr[JointIndex]] = Value;
		}
		else
		{
			UE_LOG(LogMujoco, Warning, TEXT("Failed to set joint value for index: %d"), JointIndex);
		}
	}
}

double AMujocoVolumeActor::GetJointValue(const FString& JointName) const 
{
	if (MujocoModel)
	{
		int32 JointID = mj_name2id(MujocoModel.Get(), mjOBJ_JOINT, TCHAR_TO_ANSI(*JointName));
		if (JointID >= 0)
		{
			return MujocoData->qpos[MujocoModel->jnt_qposadr[JointID]];
		}
		else
		{
			UE_LOG(LogMujoco, Warning, TEXT("Failed to get joint value for: %s"), *JointName);
		}
	}
	return 0.0;
}

double AMujocoVolumeActor::GetJointValueByIndex(int32 JointIndex) const 
{
	if (MujocoModel)
	{
		if (JointIndex >= 0 && JointIndex < MujocoModel->njnt)
		{
			return MujocoData->qpos[MujocoModel->jnt_qposadr[JointIndex]];
		}
		else
		{
			UE_LOG(LogMujoco, Warning, TEXT("Failed to get joint value for index: %d"), JointIndex);
		}
	}
	return 0.0;
}

void AMujocoVolumeActor::PostRegisterAllComponents()
{
	Super::PostRegisterAllComponents();

	if (GetWorld())
	{
		Options->ModelName = GetWorld()->GetMapName();		
	}
}

void AMujocoVolumeActor::Tick(float DeltaTime) 
{
	if (MujocoData)
	{
		mj_step(MujocoModel.Get(), MujocoData.Get());

		for (int32 Frame = 0; Frame < FrameSkip; ++Frame)
		{
			mj_step(MujocoModel.Get(), MujocoData.Get());
		}

		for (int32 i = 1; i < BodyComponents.Num(); ++i)
		{
			if (BodyComponents[i].IsValid())
			{
				mjtNum* Pos = MujocoData->xpos + 3 * i;
				mjtNum* Quat = MujocoData->xquat + 4 * i;

				FVector Position = FVector(Pos[0], -Pos[1], Pos[2]) * 100.0f;
				FQuat Rotation = FQuat(Quat[1], -Quat[2], Quat[3], -Quat[0]);

				BodyComponents[i]->SetWorldLocationAndRotation(Position, Rotation);
				BodyComponents[i]->UpdateBounds();
				BodyComponents[i]->UpdateOverlaps();
			}
			else
			{
				UE_LOG(LogMujoco, Warning, TEXT("Invalid body component at index: %d"), i);
			}
		}
	}	
}

void AMujocoVolumeActor::PostInitializeComponents()
{
	Super::PostInitializeComponents();
}

void AMujocoVolumeActor::BeginPlay()
{
	Super::BeginPlay();	
}

void AMujocoVolumeActor::EndPlay(const EEndPlayReason::Type EndPlayReason)
{
	Super::EndPlay(EndPlayReason);

	MujocoData.Reset();
	MujocoModel.Reset();
}
